/*
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.cc.CcFile
 *
 * (C) Copyright IBM Corporation 2004, 2013.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.wvcm.stp.cc;

import static com.ibm.rational.wvcm.stpex.StpExBase.PROPERTY_NAMESPACE;

import java.io.File;
import java.io.OutputStream;
import java.util.List;

import javax.wvcm.ControllableResource;
import javax.wvcm.Feedback;
import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.Resource;
import javax.wvcm.WvcmException;

import com.ibm.rational.wvcm.stp.cc.CcVersion.CcMergeFlag;
import com.ibm.rational.wvcm.stpex.StpExEnumeration;

/**
 * A proxy for a file, directory, or symbolic link resource in a ClearCase view.
 * This resource is either under version control or could potentially be
 * put under version control.
 */
public interface CcFile
    extends CcResource, ControllableResource
{
    /** Flags for the doApplyAttribute and doRemoveAttribute methods */
    enum AttributeOpFlag implements StpExEnumeration {

        /**
         * Replace existing attribute instance.
         * (Only applies to doApplyAttribute)
         */
        REPLACE("replace"),

        /**
         * Apply/remove attribute recursively over sub-tree.
         */
        RECURSE("recurse"),

        /**
         * If the attribute type was created with a default value, uses
         * that value for the attribute instead of the value specified in
         * the call. An error occurs if the attribute type was not created 
         * with a default value. 
         * (Only applies to doApplyAttribute)
         */
        DEFAULT("default");
        
        private String m_name;

        private AttributeOpFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }


    /** Flags for the doApplyLabel method */
    enum ApplyLabelFlag implements StpExEnumeration {

        /**
         * Replace existing label instance.
         */
        REPLACE("replace"),

        /**
         * Apply label recursively over sub-tree.
         */
        RECURSE("recurse"),

        /**
         * Follow symbolic links
         */
        FOLLOW_SYMLINKS("follow-symlinks");

        private String m_name;

        private ApplyLabelFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }
    
    /** Flags for the doRemoveLabel method */
    enum RemoveLabelFlag implements StpExEnumeration {

        /**
         * Remove label recursively over sub-tree.
         */
        RECURSE("recurse"),

        /**
         * Follow symbolic links
         */
        FOLLOW_SYMLINKS("follow-symlinks");

        private String m_name;

        private RemoveLabelFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }

    /** Flags for the doUncheckout method */
    enum UncheckoutFlag implements StpExEnumeration {
        /**
         * Preserve changes in checked out version in a ".keep" file.
         */
        KEEP("keep");

        private String m_name;

        private UncheckoutFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }

    /** Flags for the doUnhijack method */
    enum UnhijackFlag implements StpExEnumeration {

        /**
         * Preserve changes in hijacked version in a ".keep" file.
         */
        KEEP("keep");

        private UnhijackFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }

        private String m_name;
    }

    /**
     * Flags for the <code>doRefresh</code> and <code>doRestore</code> methods.
     */
    enum RefreshFlag implements StpExEnumeration {

        /**
         * Do <i>not</i> refresh hijacked files.
         * Leave hijacked files in the hijacked state, and do not alter
         * their contents.
         * <p>
         * Note: a deleted file or directory is considered to be hijacked.
         * In order to refresh or restore a deleted file or directory,
         * you must specify <code>OVERWRITE_HIJACKS</code> or
         * <code>RENAME_HIJACKS</code>.
         * </p>
         * This is the default hijack behavior for both <code>doRefresh</code>
         * and <code>doRestore</code>.
         */
        KEEP_HIJACKS("keep-hijacks"),

        /**
         * If a file being refreshed is hijacked in this view,
         * overwrite the hijacked contents with the new version's contents.
         * Do not preserve the hijacked contents.
         */
        OVERWRITE_HIJACKS("overwrite-hijacks"),

        /**
         * If a file being refreshed is hijacked in this view,
         * preserve the hijacked contents by moving the hijacked file to
         * <code><file-name>.keep</code>.
         */
        RENAME_HIJACKS("rename-hijacks"),

        /**
         * When refreshing a file to a different version, set the file's
         * "last modified" time to be the time when the version was created.
         * By default, a refreshed file's "last modified" time will simply be
         * the time when the <code>doRefresh</code> was performed.
         */
        PRESERVE_CREATION_TIME("preserve-creation-time"),

        /**
         * Preview the refresh operation, but don't actually perform it.
         * Invoke the caller's feedback methods to inform them what the
         * refresh would do if it were performed right now.
         */
        PREVIEW_ONLY("preview-only");

        private String m_name;

        private RefreshFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }
    
    /** Flags for the <code>doCheckout</code> method. */
    enum CcCheckoutFlag implements StpExEnumeration {
        
        /**
         * Indicates whether to checkout this file reserved.
         */
        RESERVED("reserved"),
        
        /**
         * Fall back to unreserved if a reservation can not be obtained.
         */
        FALLBACK_TO_UNRESERVED("fallback-to-unreserved"),
        
        /**
         * Checkout the version of the file that is currently loaded in the
         * view, even if that version is not the version selected by the
         * view's config spec.
         * 
         * <p>If the loaded version is not the version selected by the view's
         * config spec, and neither {@link #LOADED_NOT_LATEST} nor
         * {@link #LATEST_NOT_LOADED} flags are set, the checkout operation
         * will throw an exception with reason code <code>CHECKOUT_NOT_LATEST</code>.
         */
        LOADED_NOT_LATEST("checkout-loaded-not-latest"),
        
        /**
         * Checkout the version of the file that is selected by the view's 
         * config spec. If this version is not loaded at checkout time, then 
         * load it prior to performing the checkout.
         * 
         * <p>If the loaded version is not the version selected by the view's
         * config spec, and neither {@link #LOADED_NOT_LATEST} nor
         * {@link #LATEST_NOT_LOADED} flags are set, the checkout operation
         * will throw an exception with reason code <code>CHECKOUT_NOT_LATEST</code>.
         */
        LATEST_NOT_LOADED("checkout-latest-not-loaded"),
        
        /**
         * Indicates whether to checkout this file even if the currently
         * selected branch is mastered by another replica. The
         * <code>RESERVED</code> flag must NOT be set with this flag.
         * 
         * <p>If the file is mastered by this replica, setting this
         * flag has no effect.
         */
        NON_MASTERED_OK("non-mastered-ok"),
        
        /**
         * If the file is hijacked, keep the hijacked contents upon checkout.
         */
        PRESERVE_HIJACK_CONTENTS("preserve-hijack-contents"),
        
        /**
         * After a file is 'checkedout', set the file's "last modified" 
         * time to be the time when the version was first created.
         * <p>This applies only to dynamic views.
         */
        PRESERVE_MODIFICATION_TIME("preserve-modification-time");  
        
        private String m_name;

        private CcCheckoutFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }

    /** Flags for the <code>doCcVersionControl</code> method. */
    enum CcVersionControlFlag implements StpExEnumeration {
        
        /**
         * Indicates whether to checkin this file after it is added to version control.
         * The default is to leave it checked out.<p>
         * This flag is mutually exclusive with {@link CcVersionControlFlag#NO_CHECKOUT}.
         */
        CHECKIN("checkin"),
        
        /**
         * Do not checkout the file after adding to version control.<p>
         * This flag is mutually exclusive with {@link CcVersionControlFlag#CHECKIN}.
         */
        NO_CHECKOUT("no-checkout"),
        
        /**
         * Assigns mastership of the <b>main</b> branch of the newly version controlled
         * file to the VOB replica in which you execute operation.  By default
         * mastership of the file's <b>main</b> branch is assigned to the VOB replica 
         * that masters the <b>main</b> branch type. <p>
         * This flag is mutually exclusive with {@link CcVersionControlFlag#MAKE_PATH}.
         */
        MASTER_LOCALLY("master-locally"),
        
        /**
         * Automatically checkout the version controlled parent directory and
         * check back in after new file has been added to version control.
         * Any view private parent directories below the version controlled parent
         * and the desired file, will also be version controlled.<p>
         * This flag is mutually exclusive with {@link CcVersionControlFlag#MASTER_LOCALLY}.
         */
        MAKE_PATH("mkpath");
        
        private String m_name;

        private CcVersionControlFlag(String name) { m_name = name; }

        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString() { return m_name; }
    }

    /** Values for file or directory load state */
    enum LoadState implements StpExEnumeration {

        /**
         * This file or directory is loaded in its file area.
         */
        LOADED,

        /**
         * This directory is partially loaded in its file area, i.e.,
         * some but not all of its children are loaded.
         */
        PARTIALLY_LOADED,

        /**
         * This file or directory is not loaded in its file area.
         */
        NOT_LOADED;
    }

    /**
     * Create a new view-private file at the location specified by this resource.
     * The request will fail if a resource already exists at that location.
     * @param feedback 
     * @return a CcFile proxy for the new file
     * @see javax.wvcm.ControllableResource#doCreateResource(Feedback)
     */
    public CcFile createCcFile(Feedback feedback) throws WvcmException;
    
    /**
     * Apply the specified attribute to the checked-in version of this controllable resource.
     * @param flags array of flags which specify the behavior of the operation
     * @param comment Comment (if any) to be used for operation.  Empty string if none.
     * @param attributeName Name of an existing attribute type to be used to create 
     * an instance will to be applied to this resource.
     * @param attributeValue Value of attribute instance.  If the vtype of the attribute type is
     * a string, it must be enclosed in additional quotes <em>within the string</em>.  For example, if
     * specified as a constant it would appear as <code>"\"string value\""</code>.  If the
     * vtype is not a string, this should be a string representation of the given value 
     * (e.g. <code>"3.1415"</code>, <code>"0xa413"</code>, etc.).
     * @param versionId Applies the attribute to the explicitly specified version,
     * overriding the version selected by the view. 
     * This string must only represent the version suffix (e.g. /main/314).
     * @param feedback 
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException
     */
    ControllableResource doApplyAttribute(AttributeOpFlag[] flags, String comment, 
            String attributeName, String attributeValue, String versionId, Feedback feedback)
    throws WvcmException;
    
    /**
     * Remove the specified attribute from the checked-in version of this controllable resource.
     * @param flags array of flags which specify the behavior of the operation
     * @param comment Comment (if any) to be used for operation.  Empty string if none.
     * @param attributeName Name of the attribute to be removed from this resource
     * @param versionId Removes the attribute from the explicitly specified version,
     * overriding the version selected by the view.
     * @param feedback 
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException
     */
    ControllableResource doRemoveAttribute(AttributeOpFlag[] flags, String comment, 
            String attributeName, String versionId, Feedback feedback)
    throws WvcmException;

    /**
     * Apply the specified label to the checked-in version of this controllable resource.
     * @param flags array of flags which specify the behavior of the operation    
     * @param label the label to be added to this version
     * @param feedback 
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException
     */
    ControllableResource doApplyLabel(ApplyLabelFlag[] flags,
            String label, Feedback feedback)
        throws WvcmException;    
    
    /**
     * Apply the specified label to the checked-in version of this controllable resource.
     * @param flags array of flags which specify the behavior of the operation
     * @param comment The comment for this operation, or null for no comment   
     * @param label the label to be added to this version
     * @param feedback 
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException
     */
    ControllableResource doApplyLabel(ApplyLabelFlag[] flags, String comment,
            String label, Feedback feedback)
        throws WvcmException;
    
    /**
     * Remove the specified label from the checked-in version of this controllable resource.
     * @param flags array of flags which specify the behavior of the operation     
     * @param label the label to be removed from this version
     * @param feedback 
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException
     */
    ControllableResource doRemoveLabel(RemoveLabelFlag[] flags,
            String label, Feedback feedback)
        throws WvcmException;
    
    /**
     * <p>Check out this version-controlled file or directory resource.
     * The resource is checked out to the ClearCase view implied by its location.
     * </p>
     * <p>If the view is a UCM view, the caller must insure there is a
     * {@link javax.wvcm.Workspace#CURRENT_ACTIVITY} for this operation.
     * The checked out file will be added to the current activity's change set.
     * The caller may explicitly specify an activity using this resource's
     * {@link javax.wvcm.ControllableResource#setActivity} method.  In that case,
     * the specified activity will become the new current activity.
     * Otherwise, the existing current activity will be used.
     * If the view is a UCM view and there is no current activity, the operation
     * will fail.
     * </p>
     * <p>The caller may optionally specify a checkout comment using this
     * resource's {@link javax.wvcm.Resource#setComment} method.
     * </p>
     * @param flags array of flags which specify the behavior of the operation
     * @param feedback 
     * @return new CcFile proxy representing the checked out file.
     * @throws WvcmException
     */
    CcFile doCcCheckout(CcCheckoutFlag[] flags, Feedback feedback)
        throws WvcmException;

    /**
     * Cancel the checkout of this version-controlled resource,
     * and restore its content to the state of its CHECKED_IN version.
     * @param flags array of flags which specify the behavior of the operation
     * @param feedback 
     * @return new CcFile proxy representing the file whose checkout has been canceled
     * @throws WvcmException
     * @see javax.wvcm.ControllableResource#doUncheckout
     */
    CcFile doUncheckout(UncheckoutFlag[] flags, Feedback feedback)
        throws WvcmException;

    /**
     * <p>Check in this checked out file or directory resource.
     * </p>
     * <p>If this resource is in a UCM view, it was added to an activity's
     * change set at checkout time.  The caller may specify an alternate
     * change set using this resource's
     * {@link javax.wvcm.ControllableResource#setActivity} method just before
     * calling <code>doCheckin</code>. 
     * </p>
     * <p>The caller may also specify a checkin comment using this
     * resource's {@link javax.wvcm.Resource#setComment} method.
     * This will override the comment specified at checkout time, if any.
     * </p>
     * @param flags array of flags which specify the behavior of the operation
     * @param feedback 
     * @return new ControllableResource proxy representing the checked in file.
     * @throws WvcmException
     * @see javax.wvcm.ControllableResource#doCheckin
     * @see com.ibm.rational.wvcm.stp.cc.CcFile#doCheckout
     */
    ControllableResource doCheckin(CheckinFlag[] flags, Feedback feedback) 
        throws WvcmException;

    /**
     * Merges the contents of two or more versions, representing files or 
     * directories, into this file. Versions must be of the same element as this
     * file.
     * 
     * @param baseVersion Base contributor. Can be null.
     * @param contribList One or more contributing versions.
     * @param flags Specify options for the merge. Can be null.
     * @param listener Callback notifier for the merge.
     * @param feedback 
     * @return new CcFile proxy representing the merged file.
     * @throws WvcmException if there is an error during the merge.
     */
    CcFile doMerge(CcVersion baseVersion,
            List<CcVersion> contribList,
            CcMergeFlag[] flags,
            CcListener listener,
            Feedback feedback) throws WvcmException;
    
    /**
     * Places the view-private file specified by this proxy under version control.
     * <p>If the view is a UCM view, the caller must insure there is a
     * {@link javax.wvcm.Workspace#CURRENT_ACTIVITY} for this operation.
     * The newly version controlled file will be added to the current activity's change set
     * and left in a checked-in state.
     * The caller may explicitly specify an activity using this resource's
     * {@link javax.wvcm.ControllableResource#setActivity} method.  In that case,
     * the specified activity will become the new current activity.
     * Otherwise, the existing current activity will be used.
     * If the view is a UCM view and there is no current activity, the operation
     * will fail.
     * </p>
     * <p>The caller may optionally specify a creation comment using this
     * resource's {@link javax.wvcm.Resource#setComment} method.
     * </p>
     * <p>The caller may optionally specify the type of element to be created using
     * this resource's {@link com.ibm.rational.wvcm.stp.cc.CcFile#setElementType} method.
     * </p>
     * @param feedback
     * @return new ControllableResource proxy representing the version controlled file.
     * @throws WvcmException
     * @see javax.wvcm.ControllableResource#doVersionControl(javax.wvcm.Feedback)
     */
    ControllableResource
    doVersionControl(Feedback feedback)
        throws WvcmException;

    /**
     * Enhanced variant of the doVersionControl that provides additional flags
     * which modify the behaviour.<p>
     * This method has two main difference from <code>doVersionControl</code>:
     * <ol>
     * <li>It does <b>not</b> automatically checkout the version controlled
     * parent directory.  The caller must do this explicitly unless the
     * {@link com.ibm.rational.wvcm.stp.cc.CcFile.CcVersionControlFlag#MAKE_PATH}
     * flag is specified.
     * <li>The newly version controlled file will be left in a checked-out state.
     * To have the resulting file checked-in, specify the
     * {@link com.ibm.rational.wvcm.stp.cc.CcFile.CcVersionControlFlag#CHECKIN} flag.
     * </ol>
     * @param flags array of flags which specify the behavior of the operation
     * @param feedback
     * @return A new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException
     * @see com.ibm.rational.wvcm.stp.cc.CcFile#doVersionControl(javax.wvcm.Feedback)
     */
    ControllableResource
    doCcVersionControl(CcVersionControlFlag flags[], Feedback feedback)
        throws WvcmException;
    
    /**
     * <p>
     * Refresh this file or directory. Re-evaluate the
     * view's config spec and update resources on the client to be whatever
     * is currently selected by the config spec.  If this is a directory,
     * recursively refresh its contents.
     * </p>
     * <p>
     * Example: The config spec says "element * /main/LATEST", and you
     * have version "/main/7" of this resource loaded. Someone else checks
     * in a new version, thus creating a "/main/8". In that case,
     * doRefresh() will cause version "/main/8" of this resource to
     * be loaded into your view.
     * </p>
     * <p>
     * Preconditions: This resource must be a version-controlled file
     * or directory.
     * </p>
     * <p>
     * Postconditions: This resource (and its descendants if this is a
     * directory) are updated to be what is currently selected by the view's
     * config spec.
     * </p>
     * @param flags array of flags which specify the behavior of the operation
     * @param feedback a list of properties to fetch on the
     * updated resources. The resources returned by the iterator returned
     * by doRefresh will have these properties filled in.
     * @return a new CcFile proxy representing the refreshed file
     * @throws WvcmException if the precondition is not met or if an error
     */
    CcFile doRefresh(RefreshFlag[] flags, Feedback feedback)
       throws WvcmException;

    /**
     * <p>
     * Refresh this file or directory. Re-evaluate the
     * view's config spec and update resources on the client to be whatever
     * is currently selected by the config spec.  If this is a directory,
     * recursively refresh its contents.
     * </p>
     * <p>
     * Example: The config spec says "element * /main/LATEST", and you
     * have version "/main/7" of this resource loaded. Someone else checks
     * in a new version, thus creating a "/main/8". In that case,
     * doRefresh() will cause version "/main/8" of this resource to
     * be loaded into your view.
     * </p>
     * <p>
     * Preconditions: This resource must be a version-controlled file
     * or directory.
     * </p>
     * <p>
     * Postconditions: This resource (and its descendants if this is a
     * directory) are updated to be what is currently selected by the view's
     * config spec.
     * </p>
     * @param flags array of flags which specify the behavior of the operation
     * @param listener Callback notifier for the merge.
     * @param feedback a list of properties to fetch on the
     * updated resources. The resources returned by the iterator returned
     * by doRefresh will have these properties filled in.
     * @return a new CcFile proxy representing the refreshed file
     * @throws WvcmException if the precondition is not met or if an error
     */
    CcFile doCcRefresh(RefreshFlag[] flags, CcListener listener, Feedback feedback)
    throws WvcmException;
    
    /**
     * <p>
     * Restore this file or directory. If this is a directory, recursively
     * restore its contents. This repairs any damage to the portion of the file
     * area database specified by this resource.
     * </p>
     * 
     * @param flags array of flags which specify the behavior of the operation
     * @param feedback
     * @return a new CcFile proxy representing the restored file
     * @throws WvcmException
     */
    CcFile doRestore(RefreshFlag[] flags, Feedback feedback)
       throws WvcmException;

    /**
     * <p>
     * Load this controllable resource into the web view's local file area.
     * If this is a controllable folder, loads the tree of controllable
     * resources under this folder. 
     * Also updates the web view's load rules if necessary
     * to include this file or folder.
     * </p>
     * <p>
     * If this resource was already loaded, doLoad is a no-op.
     * </p>
     * <p>
     * Preconditions: This must be a version-controlled file or folder
     * in a web view.
     * </p>
     * <p>
     * Postconditions: This file, or the tree of files under this folder,
     * is loaded into the web view. Thus, the file(s) appear, and the view's
     * load rules are updated to include this file or folder.
     * </p>
     * @param feedback (optional) feedback's notifyIsModified() method will be
     *           called for each file or directory as it is loaded
     * @return a new CcFile proxy for the file that has been loaded
     * @throws WvcmException
     */
    CcFile doLoad(Feedback feedback) throws WvcmException;

    /**
     * Hijack this web view controllable resource.
     * Make the resource writable and set its "last modified" time to the
     * current time.  This operation does <i>not</i> contact the server.
     * @param feedback 
     * @return a new CcFile proxy for the hijacked file
     * @throws WvcmException
     */
    CcFile hijack(Feedback feedback)
       throws WvcmException;
    
    /**
     * Undo a hijack on this hijacked controllable resource.  Reload the file's
     * contents from the appropriate version on the server.
     * @param flags Specify <code>KEEP</code> to save a copy of the hijacked
     * file's contents in a ".keep" file before undoing the hijack.
     * @param feedback 
     * @return a new CcFile proxy for the unhijacked file
     * @throws WvcmException
     */
    CcFile doUnhijack(UnhijackFlag[] flags, Feedback feedback)
       throws WvcmException;

    /**
      * For client resources, identifies the file system path to the local file
      * representing this controllable resource.
      */
     PropertyName<File> CLIENT_PATH =
         new PropertyName<File>(PROPERTY_NAMESPACE, "client-path");

     /**
      * Returns the value of this proxy's {@link #CLIENT_PATH} property.
      * @return The path to this ControllableResource in the local file area.
      *         Will never be <code>null</code>.
      * @throws WvcmException if requested of a ControllableResource without client state
      */
     File getClientPath() throws WvcmException;

     /**
      * For version-controlled resources, this resource's element type.
      */
     PropertyName<CcElementType> ELEMENT_TYPE =
         new PropertyName<CcElementType>(PROPERTY_NAMESPACE, "element-type");

     /**
      * Returns the value of this proxy's {@link #ELEMENT_TYPE} property.
      * @return This resource's element type.  Will be <code>null</code>
      *         if resource is not version-controlled.
     * @throws WvcmException  if this proxy doesn't define a value for this property.
      */
     CcElementType getElementType() throws WvcmException;

     /**
      * Set the value of this file or directory's {@link #ELEMENT_TYPE} property.
      * This property can only be set just prior to calling doVersionControl()
      * on the resource.
      * @param eltype resource's desired element type
      */
     void setElementType(CcElementType eltype);

     /**
     * Is this file a web view file area database file?  File area DB files require
     * special treatment.  For instance, they cannot be source controlled.
     * For this reason, applications may choose to hide these files from
     * the end user.
     */
    PropertyName<Boolean> IS_DB_FILE =
        new PropertyName<Boolean>(PROPERTY_NAMESPACE, "is-db-file");

    /**
     * Get the value of this files {@link #IS_DB_FILE} property.
     * @return true if this file is a file area database file, else false
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
     */
    public boolean getIsDbFile() throws WvcmException;

    /**
     * Property which is true/false depending on whether this version-controlled 
     * resource has been hijacked.
     */
    PropertyName<Boolean> IS_HIJACKED =
        new PropertyName<Boolean>(PROPERTY_NAMESPACE, "is-hijacked");

    /**
     * Get the value of this file's {@link #IS_HIJACKED} property.
     * @return true if the file is hijacked, false otherwise
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
    */
    public boolean getIsHijacked() throws WvcmException;

    /**
     * Is this file actually a symbolic link?
     */
    PropertyName<Boolean> IS_SYMLINK =
        new PropertyName<Boolean>(PROPERTY_NAMESPACE, "is-symlink");

    /**
     * Get the value of the {@link #IS_SYMLINK} property.
     * @return true if this resource is a symbolic link, false otherwise.
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
    */
    public boolean getIsSymlink() throws WvcmException;

    /**
     * Returns the direct parent of this file.  The value will
     * be <code>null</code> if the file has no parent (e.g. VOB root).
     * Does not find parents of hard-linked names for the file.
     */
    public static final PropertyName<CcFile> PARENT =
        new PropertyName<CcFile>(PROPERTY_NAMESPACE, "cc-parent");
    
    /**
     * Get the value of the {@link #PARENT} property.
     * @return A CcFile proxy for the parent, null if no parent
     * @throws WvcmException
     */
    public CcFile getParent() throws WvcmException;

    /**
     * <p>If this file is actually a symbolic link, the path to its target file
     * or directory.  The path is interpreted relative to this file's parent 
     * directory.</p>
     * 
     * <p>If this file is not a symbolic link, this value will be 
     * <code>null</code>.</p>
     */
    PropertyName<String> SYMLINK_TARGET_PATH =
        new PropertyName<String>(PROPERTY_NAMESPACE, "symlink-target-path");

    /**
     * Get the value of this resource's {@link #SYMLINK_TARGET_PATH} property.
     * @return path to symlink's target file or directory, or <code>null</code>
     *         if this file is not a symbolic link.
     * @throws WvcmException if property was not requested
     */
    public String getSymlinkTargetPath() throws WvcmException;

    /**
     * The view-relative path for this resource.
     */
    PropertyName<String> VIEW_RELATIVE_PATH =
        new PropertyName<String>(PROPERTY_NAMESPACE, "cr-view-relative-path");

    /**
     * Get the value of this resource's {@link #VIEW_RELATIVE_PATH} property.
     * @return view-relative path
     * @throws WvcmException if property was not requested
     */
    public String getViewRelativePath() throws WvcmException;

    /**
     * The CC version resource associated with this file.
     * The value of this property will be null if this is not a version-
     * controlled resource.
     * @see javax.wvcm.ControllableResource#CHECKED_IN and 
     * javax.wvcm.ControllableResource#CHECKED_OUT
     */
    PropertyName<CcVersion> VERSION =
        new PropertyName<CcVersion>(PROPERTY_NAMESPACE, "version");

    /**
     * Get the value of this resource's {@link #VERSION} property.
     * @return this file's version, or <code>null</code> if this file is
     * not version controlled
     * @throws WvcmException if property was not requested
     */
    public CcVersion getVersion() throws WvcmException;

    /**
     * The OID of the CC version resource associated with this file.
     * The value of this property will be null if this is not a version-
     * controlled resource.
     */
    PropertyName<String> VERSION_OID =
        new PropertyName<String>(PROPERTY_NAMESPACE, "version-oid");

    /**
     * Get the value of this resource's {@link #VERSION_OID} property.
     * @return this file's version oid, or <code>null</code> if this file is
     * not version controlled
     * @throws WvcmException if property was not requested
     */
    public String getVersionOid() throws WvcmException;

    /**
     * The tag of the VOB in which this file resides.
     */
    PropertyName<CcVobTag> VOB_TAG =
        new PropertyName<CcVobTag>(PROPERTY_NAMESPACE, "file-vob-tag");
 
    /**
     * Get the value of this resource's {@link #VOB_TAG} property.
     * @return the VOB tag for this file's VOB as a CcVobTag proxy, 
     * or <code>null</code> if this file is not version controlled
     * @throws WvcmException if property was not requested
     */
    public CcVobTag getVobTag() throws WvcmException;
    
    /**
     * The CC element resource associated with this file.
     * The value of this property will be null if this is not a version-
     * controlled resource.
     * @see javax.wvcm.ControllableResource#VERSION_HISTORY
     */
    PropertyName<CcElement> ELEMENT =
        new PropertyName<CcElement>(PROPERTY_NAMESPACE, "element");

    /**
     * Get the value of this resource's {@link #ELEMENT} property.
     * @return this file's element, or <code>null</code> if this file is
     * not version controlled
     * @throws WvcmException if property was not requested
     */
    public CcElement getElement() throws WvcmException;

    /**
     * Returns a {@link java.io.File File} representing the location of this
     * client-side resource in the local file system, else <code>null</code> if
     * this resource is a server-side resource only, e.g., if it is a proxy
     * to an unloaded file in a web view.
     * @return The location of this resource in the local file system, else
     * <code>null</code> if this is a proxy to a server-side resource only.
     * 
     * @throws WvcmException Thrown if there was an error in determining the
     * path for this client-side Resource.
     */
    File clientResourceFile() throws WvcmException;
    
    /**
     * Reads the file's properties, if they are available locally on the client machine.
     * @param feedback the properties to be available in the returned proxy,
     *  and any other feedback desired, such as progress indications.
     * @return a {@link CcFile} containing the wanted properties
     * that are available locally on the client machine
     * without communicating with the server.
     * @see Resource#doReadProperties(Feedback) doReadProperties.
     * @throws WvcmException ReasonCode:
     * <br> {@link javax.wvcm.WvcmException.ReasonCode#NOT_FOUND}:
     *  The file MUST be available locally on the client machine.
     */
    CcFile readProperties(Feedback feedback) throws WvcmException;

    /**
     * Reads the file's content, if it is available locally on the client machine.
     * @see Resource#doReadContent(OutputStream,Feedback) doReadContent
     * @param feedback the properties to be available in the returned proxy,
     *  and any other feedback desired, such as progress indications.
     * @param content the file's content, if available, is written to this
     * byte stream.  The stream is then closed.
     * @return a CcFile proxy containing the wanted properties
     * that are available on the client host without communicating with the server.
     * @throws WvcmException ReasonCode:
     * <br> {@link javax.wvcm.WvcmException.ReasonCode#NOT_FOUND}:
     *  The file MUST be available locally on the client machine.
     */
    CcFile readContent(OutputStream content, Feedback feedback) throws WvcmException;

    /**
     * Resolve this file resource, but do not consult the ClearCase server.
     * Unlike {@link CcResource#doResolve()}, use only information currently
     * available information on the local client machine.  
     * @see CcResource#doResolve()
     * @return a new file proxy of the correct, most specific resource type
     * @throws WvcmException with NOT_FOUND reason code if this file does not
     *         exist on the local machine.
     */
    CcFile resolve() throws WvcmException;
    
    /**
     * Whether this file is loaded, partially loaded, or not loaded in the
     * web view file area in which it resides.
     */
    PropertyName<LoadState> LOAD_STATE =
        new PropertyName<LoadState>(PROPERTY_NAMESPACE, "load-state");

    /**
     * Get the value of this resource's {@link #LOAD_STATE} property.
     * @return LoadState.LOADED, LoadState.PARTIALLY_LOADED or LoadState.NOT_LOADED
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
     */
    LoadState getLoadState() throws WvcmException;

    /**
     * <p>Version controlled files and directories in a web view
     * may have both client state -
     * state maintained in a local file area - and server state.  When the
     * client state and server state get out of sync, we call this <i>skew</i>.
     * </p>
     * <p>By definition, skew cannot occur in a dynamic view.
     * </p>
     * <p>The caller can detect skew by including the {@link #SKEWED_PROPERTY_LIST}
     * property in property requests to <code>doReadProperties()</code> and
     * other <code>do</code> methods that accept property requests.
     * 
     * The resulting value of this property is the list of property names
     * in the request that differed between the client and the server for this
     * particular file or directory.
     * </p>
     * <p>Note that only certain properties are checked for skew - properties
     * where skew can cause significant problems.  For example,
     * {@link #IS_CHECKED_OUT}, {@link #IS_VERSION_CONTROLLED}, and
     * {@link #VERSION_OID}.
     * 
     * Note that skew can also be caused by elementness skew (file vs dir) and
     * loadness skew (loaded vs unloaded).
     * </p>
     * <p>Note also that <i>only</i> properties in the current property request
     * are checked for skew.
     * </p>
     * <p>NOTE: This should be used only as a trigger to do a real discordance
     * scan of the directory. These values are not reliable enough to use for
     * icon decoration or user messages. This is a quick and easy way to 
     * automatically detect skew, but it does not cover all edge cases 
     * (symlinks, aliases, and others) or discordance types.
     * </p>
     */
    PropertyName<List<PropertyName>> SKEWED_PROPERTY_LIST =
        new PropertyName<List<PropertyName>>(PROPERTY_NAMESPACE, "skewed-property-list");

    /**
     * Get the value of this file's {@link #SKEWED_PROPERTY_LIST} property.
     * 
     * NOTE: This should be used only as a trigger to do a real discordance
     * scan of the directory. These values are not reliable enough to use for
     * icon decoration or user messages. This is a quick and easy way to 
     * automatically detect skew, but it does not cover all edge cases 
     * (symlinks, aliases, and others) or discordance types.
     * 
     * @return a list of property names in the current property request whose
     *         values differed between the client and the server for this file
     *         or directory.
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
     */
    List<PropertyName> getSkewedPropertyList() throws WvcmException;
    
    /**
     * The config spec rule that selects this file.
     */
    PropertyName<String> SELECTION_RULE =
        new PropertyName<String>(PROPERTY_NAMESPACE, "selection-rule");
    
    /**
     * Get the value of this resource's {@link #SELECTION_RULE} property.
     * @return this file's config spec rule, as a string.
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
     */
    String getSelectionRule() throws WvcmException;
    
    /**
     * The version of this file that is currently the latest on the same branch.
     */
    PropertyName<CcVersion> LATEST_VERSION_ON_BRANCH =
        new PropertyName<CcVersion>(PROPERTY_NAMESPACE, "latest-version-on-branch");
    
    /**
     * Get the value of this resource's {@link #LATEST_VERSION_ON_BRANCH} property.
     * @return the version of this file that is the latest on the same branch as the
     * version in the view.
     * @throws WvcmException
     *             if this proxy doesn't define a value for this property.
     */
    CcVersion getLatestVersionOnBranch() throws WvcmException;
}
